/* Target-dependent code for GDB on MIPS systems running FreeBSD/JunOS

   Copyright (C) 2002, 2003, 2004, 2006, 2007 Free Software Foundation, Inc.

   Portions cloned from mipsnbsd-tdep.c by fnf@specifix.com

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include "defs.h"
#include "gdbcore.h"
#include "regcache.h"
#include "regset.h"
#include "target.h"
#include "value.h"
#include "osabi.h"

#include "gdb_assert.h"
#include "gdb_string.h"

#include "fbsd-tdep.h"
#include "mipsfbsd-tdep.h"
#include "mips-tdep.h"

#include "solib-svr4.h"

/* Shorthand for some register numbers used below.  */
#define MIPS_PC_REGNUM  MIPS_EMBED_PC_REGNUM
#define MIPS_FP0_REGNUM MIPS_EMBED_FP0_REGNUM
#define MIPS_FSR_REGNUM MIPS_EMBED_FP0_REGNUM + 32

/* Core file support.  */

/* Number of registers in `struct reg' from <machine/reg.h>.  */
#define MIPSFBSD_NUM_GREGS	40 /* 38 */

/* Number of registers in `struct fpreg' from <machine/reg.h>.  */
/* FIXME: (krk) for this processor, it's soft-fp. */
#define MIPSFBSD_NUM_FPREGS	 0 /* 33 */

/* Supply register REGNUM from the buffer specified by FPREGS and LEN
   in the floating-point register set REGSET to register cache
   REGCACHE.  If REGNUM is -1, do this for all registers in REGSET.  */

static void
mipsfbsd_supply_fpregset (const struct regset *regset,
			  struct regcache *regcache,
			  int regnum, const void *fpregs, size_t len)
{
#if 0
  size_t regsize = mips_isa_regsize (get_regcache_arch (regcache));
  const char *regs = fpregs;
  int i;

  gdb_assert (len >= MIPSFBSD_NUM_FPREGS * regsize);

  for (i = MIPS_FP0_REGNUM; i <= MIPS_FSR_REGNUM; i++)
    {
      if (regnum == i || regnum == -1)
	regcache_raw_supply (regcache, i,
			     regs + (i - MIPS_FP0_REGNUM) * regsize);
    }
#endif
}

/* Supply register REGNUM from the buffer specified by GREGS and LEN
   in the general-purpose register set REGSET to register cache
   REGCACHE.  If REGNUM is -1, do this for all registers in REGSET.  */

static void
mipsfbsd_supply_gregset (const struct regset *regset,
			 struct regcache *regcache, int regnum,
			 const void *gregs, size_t len)
{
  size_t abisize = mips_abi_regsize (get_regcache_arch (regcache));
  const char *regs = gregs;
  int i;

  gdb_assert (len >= MIPSFBSD_NUM_GREGS * abisize);

  for (i = 0; i <= MIPSFBSD_NUM_GREGS; i++)
    {
      if (regnum == i || regnum == -1)
	{
	  size_t isasize = register_size (current_gdbarch, i);
	  void *addr = regs + i * abisize;
	  gdb_byte buf[MAX_REGISTER_SIZE];
	  LONGEST val;

	  val = extract_signed_integer (addr, abisize);
	  store_signed_integer (buf, isasize, val);
	  regcache_raw_supply (regcache, i, buf);
	}
    }

  if (len >= (MIPSFBSD_NUM_GREGS + MIPSFBSD_NUM_FPREGS) * abisize)
    {
      regs += MIPSFBSD_NUM_GREGS * abisize;
      len -= MIPSFBSD_NUM_GREGS * abisize;
      mipsfbsd_supply_fpregset (regset, regcache, regnum, regs, len);
    }
}

/* FreeBSD/mips register sets.  */

static struct regset mipsfbsd_gregset =
{
  NULL,
  mipsfbsd_supply_gregset
};

static struct regset mipsfbsd_fpregset =
{
  NULL,
  mipsfbsd_supply_fpregset
};

/* Return the appropriate register set for the core section identified
   by SECT_NAME and SECT_SIZE.  */

static const struct regset *
mipsfbsd_regset_from_core_section (struct gdbarch *gdbarch,
				   const char *sect_name, size_t sect_size)
{
  size_t regsize = mips_abi_regsize (gdbarch);

  if (strcmp (sect_name, ".reg") == 0
      && sect_size >= MIPSFBSD_NUM_GREGS * regsize)
    return &mipsfbsd_gregset;
#if 0	/* MVS/krk: no floating point */
  if (strcmp (sect_name, ".reg2") == 0
      && sect_size >= MIPSFBSD_NUM_FPREGS * regsize)
    return &mipsfbsd_fpregset;
#endif
  return NULL;
}

/* Conveniently, GDB uses the same register numbering as the
   ptrace register structure used by FreeBSD/mips.  */

void
mipsfbsd_supply_reg (char *regs, int regno)
{
  int i;

  for (i = 0; i <= PC_REGNUM; i++)
    {
      if (regno == i || regno == -1)
	{
	  if (CANNOT_FETCH_REGISTER (i))
	    regcache_raw_supply (current_regcache, i, NULL);
	  else
	    {
	      /* Convert between abi (input) and isa (output).  
		 Required in case abi regsize != isa regsize.  */
	      size_t abisize = mips_abi_regsize (current_gdbarch);
	      size_t isasize = register_size (current_gdbarch, i);
	      void *addr = regs + i * abisize;
	      gdb_byte buf[MAX_REGISTER_SIZE];
	      LONGEST val;

	      val = extract_signed_integer (addr, abisize);
	      store_signed_integer (buf, isasize, val);
	      regcache_raw_supply (current_regcache, i, buf);
	    }
        }
    }
}

void
mipsfbsd_fill_reg (char *regs, int regno)
{
  int i;

  for (i = 0; i <= PC_REGNUM; i++)
    if ((regno == i || regno == -1) && ! CANNOT_STORE_REGISTER (i))
      {
	/* Convert between isa (input) and abi (output).  
	   Required in case abi regsize != isa regsize.  */
	size_t abisize = mips_abi_regsize (current_gdbarch);
	size_t isasize = register_size (current_gdbarch, i);
	void *addr = regs + i * abisize;
	gdb_byte buf[MAX_REGISTER_SIZE];
	LONGEST val;

	regcache_raw_collect (current_regcache, i, buf);
	val = extract_signed_integer (buf, isasize);
	store_signed_integer (addr, abisize, val);
      }
}

void
mipsfbsd_supply_fpreg (char *fpregs, int regno)
{
  int i;
  size_t regsize = mips_isa_regsize (current_gdbarch);

  for (i = FP0_REGNUM;
       i <= mips_regnum (current_gdbarch)->fp_implementation_revision;
       i++)
    {
      if (regno == i || regno == -1)
	{
	  if (CANNOT_FETCH_REGISTER (i))
	    regcache_raw_supply (current_regcache, i, NULL);
	  else
            regcache_raw_supply (current_regcache, i,
				 fpregs + ((i - FP0_REGNUM) * regsize));
	}
    }
}

void
mipsfbsd_fill_fpreg (char *fpregs, int regno)
{
  int i;
  size_t regsize = mips_isa_regsize (current_gdbarch);

  for (i = FP0_REGNUM; i <= mips_regnum (current_gdbarch)->fp_control_status;
       i++)
    if ((regno == i || regno == -1) && ! CANNOT_STORE_REGISTER (i))
      regcache_raw_collect (current_regcache, i,
			    fpregs + ((i - FP0_REGNUM) * regsize));
}

/* (?) Under FreeBSD/mips, signal handler invocations can be
   identified by the designated code sequence that is used to return
   from a signal handler.  In particular, the return address of a
   signal handler points to the following code sequence:

	addu	a0, sp, 16
	li	v0, 295			# __sigreturn14
	syscall
   
   Each instruction has a unique encoding, so we simply attempt to
   match the instruction the PC is pointing to with any of the above
   instructions.  If there is a hit, we know the offset to the start
   of the designated sequence and can then check whether we really are
   executing in the signal trampoline.  If not, -1 is returned,
   otherwise the offset from the start of the return sequence is
   returned.  */

#define RETCODE_NWORDS	3
#define RETCODE_SIZE	(RETCODE_NWORDS * 4)

static const unsigned char sigtramp_retcode_mipsel[RETCODE_SIZE] =
{
  0x10, 0x00, 0xa4, 0x27,	/* addu a0, sp, 16 */
  0x27, 0x01, 0x02, 0x24,	/* li v0, 295 */
  0x0c, 0x00, 0x00, 0x00,	/* syscall */
};

static const unsigned char sigtramp_retcode_mipseb[RETCODE_SIZE] =
{
  0x27, 0xa4, 0x00, 0x10,	/* addu a0, sp, 16 */
  0x24, 0x02, 0x01, 0x27,	/* li v0, 295 */
  0x00, 0x00, 0x00, 0x0c,	/* syscall */
};

#if 0
static LONGEST
mipsfbsd_sigtramp_offset (struct frame_info *next_frame)
{
  CORE_ADDR pc = frame_pc_unwind (next_frame);
  const char *retcode = TARGET_BYTE_ORDER == BFD_ENDIAN_BIG
  	? sigtramp_retcode_mipseb : sigtramp_retcode_mipsel;
  unsigned char ret[RETCODE_SIZE], w[4];
  LONGEST off;
  int i;

  if (!safe_frame_unwind_memory (next_frame, pc, w, sizeof (w)))
    return -1;

  for (i = 0; i < RETCODE_NWORDS; i++)
    {
      if (memcmp (w, retcode + (i * 4), 4) == 0)
	break;
    }
  if (i == RETCODE_NWORDS)
    return -1;

  off = i * 4;
  pc -= off;

  if (!safe_frame_unwind_memory (next_frame, pc, ret, sizeof (ret)))
    return -1;

  if (memcmp (ret, retcode, RETCODE_SIZE) == 0)
    return off;

  return -1;
}
#endif

/* Figure out where the longjmp will land.  We expect that we have
   just entered longjmp and haven't yet setup the stack frame, so the
   args are still in the argument regs.  MIPS_A0_REGNUM points at the
   jmp_buf structure from which we extract the PC that we will land
   at.  The PC is copied into *pc.  This routine returns true on
   success.  */

#define NBSD_MIPS_JB_PC			(2 * 4)
#define NBSD_MIPS_JB_ELEMENT_SIZE	mips_isa_regsize (current_gdbarch)
#define NBSD_MIPS_JB_OFFSET		(NBSD_MIPS_JB_PC * \
					 NBSD_MIPS_JB_ELEMENT_SIZE)

static int
mipsfbsd_get_longjmp_target (CORE_ADDR *pc)
{
  CORE_ADDR jb_addr;
  char *buf;

  buf = alloca (NBSD_MIPS_JB_ELEMENT_SIZE);

  jb_addr = read_register (MIPS_A0_REGNUM);

  if (target_read_memory (jb_addr + NBSD_MIPS_JB_OFFSET, buf,
  			  NBSD_MIPS_JB_ELEMENT_SIZE))
    return 0;

  *pc = extract_unsigned_integer (buf, NBSD_MIPS_JB_ELEMENT_SIZE);

  return 1;
}

static int
mipsfbsd_cannot_fetch_register (int regno)
{
  return (regno == MIPS_ZERO_REGNUM
	  || regno == mips_regnum (current_gdbarch)->fp_implementation_revision);
}

static int
mipsfbsd_cannot_store_register (int regno)
{
  return (regno == MIPS_ZERO_REGNUM
	  || regno == mips_regnum (current_gdbarch)->fp_implementation_revision);
}

/* Shared library support.  */

/* (?) FreeBSD/mips uses a slightly different `struct link_map' than
   the other FreeBSD platforms.  */

static struct link_map_offsets *
mipsfbsd_ilp32_fetch_link_map_offsets (void)
{
  static struct link_map_offsets lmo;
  static struct link_map_offsets *lmp = NULL;

  if (lmp == NULL) 
    {
      lmp = &lmo;

      lmo.r_version_offset = 0;
      lmo.r_version_size = 4;
      lmo.r_map_offset = 4;
      lmo.r_ldsomap_offset = -1;

      /* Everything we need is in the first 24 bytes.  */
      lmo.link_map_size = 24;
      lmo.l_addr_offset = 4;
      lmo.l_addr_size = 4;
      lmo.l_name_offset = 8;
      lmo.l_name_size = 4;
      lmo.l_ld_offset = 12;
      lmo.l_ld_size = 4;
      lmo.l_next_offset = 16;
      lmo.l_next_size = 4;
      lmo.l_prev_offset = 20;
      lmo.l_prev_size = 4;
    }

  return lmp;
}

static struct link_map_offsets *
mipsfbsd_lp64_fetch_link_map_offsets (void)
{
  static struct link_map_offsets lmo;
  static struct link_map_offsets *lmp = NULL;

  if (lmp == NULL)
    {
      lmp = &lmo;

      lmo.r_version_offset = 0;
      lmo.r_version_size = 4;
      lmo.r_map_offset = 8;
      lmo.r_ldsomap_offset = -1;

      /* Everything we need is in the first 40 bytes.  */
      lmo.link_map_size = 48;
      lmo.l_addr_offset = 0;
      lmo.l_addr_size = 8;
      lmo.l_name_offset = 16; 
      lmo.l_name_size = 8;
      lmo.l_ld_offset = 24;
      lmo.l_ld_size = 8;
      lmo.l_next_offset = 32;
      lmo.l_next_size = 8;
      lmo.l_prev_offset = 40;
      lmo.l_prev_size = 8;
    }

  return lmp;
}



static void
mipsfbsd_init_abi (struct gdbarch_info info,
                   struct gdbarch *gdbarch)
{
  set_gdbarch_regset_from_core_section
    (gdbarch, mipsfbsd_regset_from_core_section);

  set_gdbarch_get_longjmp_target (gdbarch, mipsfbsd_get_longjmp_target);

  set_gdbarch_cannot_fetch_register (gdbarch, mipsfbsd_cannot_fetch_register);
  set_gdbarch_cannot_store_register (gdbarch, mipsfbsd_cannot_store_register);

  set_gdbarch_software_single_step (gdbarch, mips_software_single_step);

  /* (?) FreeBSD/mips has SVR4-style shared libraries.  */
  set_solib_svr4_fetch_link_map_offsets
    (gdbarch, (gdbarch_ptr_bit (gdbarch) == 32 ?
	       mipsfbsd_ilp32_fetch_link_map_offsets :
	       mipsfbsd_lp64_fetch_link_map_offsets));

  /* Enable TLS support.  */
  set_gdbarch_fetch_tls_load_module_address (gdbarch,
                                             svr4_fetch_objfile_link_map);
}


#if 0
static enum gdb_osabi
mipsfbsd_core_osabi_sniffer (bfd *abfd)
{
  if (strcmp (bfd_get_target (abfd), "freebsd-core") == 0)
    return GDB_OSABI_FREEBSD_ELF;

  warning (_("non freebsd-core file '%s'\n"), bfd_get_target (abfd));

  return GDB_OSABI_UNKNOWN;
}
#endif

void
_initialize_mipsfbsd_tdep (void)
{
  gdbarch_register_osabi (bfd_arch_mips, 0, GDB_OSABI_FREEBSD_ELF,
			  mipsfbsd_init_abi);
}
