# $Id: Makefile.gdb 579281 2013-05-01 00:22:24Z ib-builder $

CFLAGS+= -I${.CURDIR} -I${GDBDIR}

.include "Makefile.iberty.inc"
.include "Makefile.bfd.inc"
.include "Makefile.readline.inc"
.include "Makefile.opcodes.inc"
.include "Makefile.gdb.inc"

# Files that have conflicting generated .o filenames.
rename_list= bfd/corefile.c bfd/init.c readline/signals.c

# avoid functions which are duplicated across pseudo-libraries
skip_list=	\
	libiberty/xmalloc.c \
	readline/savestring.c \
	readline/support/wcwidth.c \
	readline/xmalloc.c \
	${rename_list}

Nskip_names= ${skip_list:${M_ListToSkip}}

_srcs:= \
	${_bfd_srcs} \
	${_gdb_srcs} \
	${_iberty_srcs} \
	${_opcodes_srcs} \
	${_readline_srcs}

SRCS= ${_srcs:O:u:${Nskip_names}}

_gensrcs=
.for c in ${rename_list}
_gensrcs+=${c:S,/,-,g}
${c:S,/,-,g}: $c ${GDBBVERSIONMK}
	echo '#include "$c"' > ${.TARGET}
.endfor

GENSRCS+= ${_gensrcs}
CLEANFILES+= ${_gensrcs}
SRCS+= ${GENSRCS}

.if ${SRCS:T:O} != ${SRCS:T:O:u}

dups!= echo ${SRCS:T:O} | tr ' ' '\012' | uniq -d

.BEGIN: throw_error
throw_error: .PHONY
	@echo There are some problem targets in SRCS
.for m in ${dups}
	@echo "  ${SRCS:M*$m}"
.endfor
	@echo Fix this makefile and try again; exit 1
.endif

.for stupid_gnu in \
xregcomp xre_exec xregexec xre_search xre_compile_fastmap xregerror xre_comp xre_set_syntax
CFLAGS+= -D${stupid_gnu}=${stupid_gnu:S/^x//}
.endfor

.if ${MACHINE} == "xlr" || ${MACHINE} == "octeon"
CFLAGS+= -DMIPS_USE_HW_WATCHPOINT
.endif

.if ${TARGET_ARCH} == ${MACHINE_ARCH}
_fork_child= gdb/fork-child.c
_infptrace= gdb/infptrace.c
_inftarg= gdb/inftarg.c
.endif

GENSRCS+= frame-unwind-kluge.c version.c
CLEANFILES+= frame-unwind-kluge.c.orig

frame-unwind-kluge.c: gdb/frame-unwind.c frame-unwind.diff
	cat ${.ALLSRC:M*.c} > ${.TARGET}
	patch ${.TARGET} ${.ALLSRC:M*.diff}

# gdb seems to have its own version Bison it likes. Used the
# checked in sources rather than regenerating the .c files.
.for c in ${__yyfiles_srcs}
$c: ${y:S/.c$/.y/}
	@echo 'avoiding "${YACC} ${YFLAGS} ${.IMPSRC}" to regenerate ${.TARGET}'
.endfor

ada-lex.c: ada-lex.l
	@echo 'avoiding "${LEX} -t ${LFALGS} ${.ALLSRC:M*.l}" to regenerate ${.TARGET}'
